/*
 * Compaction.cpp
 *
 *  Created on: 2. 11. 2014
 *	  Author: Tomas Pecka
 */

#include "Compaction.h"

#include <exception/AlibException.h>

#include <automaton/common/State.h>
#include <automaton/FSM/CompactNFA.h>
#include <automaton/FSM/DFA.h>

#include <stack>
#include <tuple>

namespace automaton {

namespace transform {

automaton::Automaton Compaction::convert(const automaton::Automaton& automaton) {
	automaton::Automaton* out = NULL;
	automaton.getData().Accept((void*) &out, Compaction::COMPACTION);
	automaton::Automaton res(std::move(*out));
	delete out;
	return res;
}

automaton::CompactNFA Compaction::convert(const automaton::CompactNFA& automaton)
{
	return automaton;
}

automaton::CompactNFA Compaction::convert(const automaton::DFA& automaton)
{
	automaton::CompactNFA res(automaton.getInitialState());
	res.setInputAlphabet(automaton.getInputAlphabet());

	string::LinearString path;
	path.setAlphabet(automaton.getInputAlphabet());

	std::map<automaton::State, bool> visited;
	for(const auto& state : automaton.getStates())
		visited.insert(std::make_pair(state, automaton.getInitialState() == state));

	std::stack<std::tuple<automaton::State, automaton::State, alphabet::Symbol>> stack; // state x lastFork x symbol we used to go to that state
	for(const auto& transition: automaton.getTransitionsFromState(automaton.getInitialState()))
	{
		stack.push(std::make_tuple(transition.second, automaton.getInitialState(), transition.first.second));
	}

	if(automaton.getFinalStates().count(automaton.getInitialState()) > 0)
		res.addFinalState(automaton.getInitialState());

	while(!stack.empty())
	{
		const automaton::State q = std::get<0>(stack.top());
		const automaton::State lastFork = std::get<1>(stack.top());
		const alphabet::Symbol symbol = std::get<2>(stack.top());
		stack.pop();

		path.appendSymbol(symbol);

		// only 1 child and nonfinal
		if(automaton.getTransitionsFromState(q).size() == 1 && automaton.getFinalStates().count(q) == 0 && automaton.getInitialState() != q)
		{
			const auto& transition = * automaton.getTransitionsFromState(q).begin();

			stack.push(std::make_tuple(transition.second, lastFork, transition.first.second));
		}
		else // fork or final state
		{
			res.addState(q);

			if(automaton.getFinalStates().count(q))
				res.addFinalState(q);

			res.addTransition(lastFork, path, q);
			path.setContent({});

			for(const auto& transition : automaton.getTransitionsFromState(q))
			{
				if(visited.find(transition.second)->second == false)
				{
					stack.push(std::make_tuple(transition.second, q, transition.first.second));
					visited.find(transition.second)->second = true;
				}
			}
		}
	}

	return res;
}

automaton::CompactNFA Compaction::convert(const automaton::NFA& automaton)
{
	automaton::CompactNFA res(automaton.getInitialState());
	// TODO
	return res;
}

void Compaction::Visit(void* data, const CompactNFA& automaton) const
{
	automaton::Automaton*& out = *((automaton::Automaton**) data);
	out = new automaton::Automaton(this->convert(automaton));
}

void Compaction::Visit(void* data, const DFA& automaton) const
{
	automaton::Automaton*& out = *((automaton::Automaton**) data);
	out = new automaton::Automaton(this->convert(automaton));
}

void Compaction::Visit(void*, const EpsilonNFA&) const
{
	throw exception::AlibException("Unsupported automaton type EpsilonNFA");
}

void Compaction::Visit(void*, const ExtendedNFA&) const
{
	throw exception::AlibException("Unsupported automaton type ExtendedNFA");
}

void Compaction::Visit(void*, const MultiInitialStateNFA&) const
{
	throw exception::AlibException("Unsupported automaton type MultiInitialStateNFA");
}

void Compaction::Visit(void* data, const NFA& automaton) const
{
	automaton::Automaton*& out = *((automaton::Automaton**) data);
	out = new automaton::Automaton(this->convert(automaton));
}

const Compaction Compaction::COMPACTION;

} /* namespace transform */

} /* namespace automaton */
